-----------------------------------
-----------------------------------
-- Whisp by Anea
--
-- Makes chatting fun again
-- Type /whisp in-game for options
-----------------------------------
-- core.lua
-- Main routines and functionionality
-----------------------------------

-- Create Ace2 instance
Whisp = AceLibrary("AceAddon-2.0"):new("AceEvent-2.0", "AceDB-2.0", "AceConsole-2.0", "AceHook-2.1", "FuBarPlugin-2.0")
local L = AceLibrary("AceLocale-2.2"):new("Whisp")

-- Default settings and variables
local defaultsProfile = {
  version = 0,                       -- The version we are running
  timeout = 120,                     -- How long should conversations be saved
  notifyDebug = false,               -- Shows debug messages
  hideInCombat = false,              -- Hide the chatbox tooltip when in combat
  colorIncoming = {1,0.5,1},         -- Color for incoming messages
  colorOutgoing = {0.73,0.73,1},     -- Color for outgoing messages
}
local defaultsChar = {
  linesShown = 10,                   -- How many lines to show in the tooltips
  growUp = true,                     -- Grow the tooltip upwards
  locked = true,                     -- Locks the chatbox tooltip in place
  colorBackground = {0.09, 0.09, 0.09, 0.8}  -- Color for the background of the tooltips
}

-- local variables
Whisp.currentBuild = 20080411           -- Latest build
Whisp.chatHistory = {}                  -- Chat cache
Whisp.tellTarget = nil                  -- Current tell-target
Whisp.lastSender = nil                  -- Person who last send you a tell
Whisp.editBoxFrame = {}                 -- ToolTip frame to attach to the editbox
Whisp.tooltipFrame = {}                 -- ToolTip frame to attach to fubar
Whisp.clickableTooltip = true           -- Ensures that the FuBar tooltip is clickable

----------------------------------
-----------------------------------
-- Initialisation functions

function Whisp:OnInitialize()
  self:RegisterDB("WhispDB")
  self:RegisterDefaults("profile", defaultsProfile)
  self:RegisterDefaults("char", defaultsChar)
  
  -- Initialise tooltip frames
  self.editBoxFrame = self:CreateEditBoxFrame()
  self.tooltipFrame = self:CreateTooltipFrame()
  Whisp:ToggleFrameLock()
  
  -- Setup the console and FuBar options-menu's
  self:SetupOptions()
end

function Whisp:OnEnable()
  -- Cache incoming and outgoing whispers
  self:RegisterEvent("CHAT_MSG_WHISPER", "ChatEventIncoming")
  self:RegisterEvent("CHAT_MSG_WHISPER_INFORM", "ChatEventOutgoing")

  -- Hook the chateditbox
  self:SecureHook("ChatEdit_OnShow")
  self:SecureHook("ChatEdit_UpdateHeader")
  self:SecureHook("ChatEdit_OnEscapePressed")
  
  -- Schedules
  self:ScheduleRepeatingEvent("WHISP_COLLECTGARBAGE", Whisp.CollectGarbage, 60)       -- Garbage collection
end

function Whisp:OnDisable()
  self:UnhookAll()                -- Disable all hooks
end

-----------------------------------
-----------------------------------
-- ChatEditBox Events

function Whisp:ChatEventIncoming(msg,plr)
  self:UpdateChatHistory(msg,plr, plr, true)
  self:UpdateTooltip()
  self:UpdateText()
  self:ChatEdit_OnShow()
end

function Whisp:ChatEventOutgoing(msg,plr)
  self:UpdateChatHistory(msg,plr, UnitName("player"), false)
  self:UpdateTooltip()
  self:UpdateText()
end

function Whisp:ChatEdit_OnShow()
  -- Only show when chateditbox is visible
  if ChatFrameEditBox:IsVisible() then
    -- Optionally hide when in combat
    if not (Whisp.db.profile.hideInCombat and InCombatLockdown()) then
      Whisp.editBoxFrame:Hide()
      if (ChatFrameEditBox:GetAttribute("chatType") == "WHISPER") then
        Whisp.tellTarget = Whisp:ToName(ChatFrameEditBox:GetAttribute("tellTarget"))
        if Whisp.chatHistory[Whisp.tellTarget] then
          Whisp:ShowEditBoxFrame(Whisp.tellTarget)
        end
      end
    end
  end
end

function Whisp:ChatEdit_UpdateHeader()
  Whisp:ChatEdit_OnShow()
end

function Whisp:ChatEdit_OnEscapePressed()
  Whisp.tellTarget=nil
  Whisp.editBoxFrame:Hide()
end

-----------------------------------
-----------------------------------
-- EditBox and fubar tooltip functions

function Whisp:ShowEditBoxFrame(plr)
  local frame = Whisp.editBoxFrame
  self:UpdateTooltipText(frame, plr)
  local l,t,b = frame:GetLeft(), frame:GetTop(), frame:GetBottom()
  frame:ClearAllPoints()
  if self.db.char.growUp then
    frame:SetPoint("BOTTOMLEFT",UIParent,"BOTTOMLEFT", l, b)
  else
    frame:SetPoint("TOPLEFT",UIParent,"BOTTOMLEFT", l, t)
  end
  frame:Show()
end

function Whisp:ShowTooltipFrame(plr)
  local frame = Whisp.tooltipFrame
  local tip = Whisp.tablet.tooltip
  frame:ClearAllPoints()
  local left, right, top = tip:GetLeft(), tip:GetRight(), tip:GetTop()
  local cx, cy = UIParent:GetRight()/2, UIParent:GetTop()/2
  local anchor, align = "TOP", "BOTTOM"
  if top > cy then
    anchor, align = "TOP", "BOTTOM"
    if right > cx + (cx/2) then
      anchor, align = "TOPRIGHT", "BOTTOMRIGHT"
    elseif left < cx - (cx/2) then
      anchor, align = "TOPLEFT", "BOTTOMLEFT"
    end
  else
    anchor, align = "BOTTOM", "TOP"
    if right > cx + (cx/2) then
      anchor, align = "BOTTOMRIGHT", "TOPRIGHT"
    elseif left < cx - (cx/2) then
      anchor, align = "BOTTOMLEFT", "TOPLEFT"
    end
  end
  frame:SetPoint(anchor, tip, align)
  self:UpdateTooltipText(frame, plr)
  frame:Show()
end

-- Update
function Whisp:UpdateTooltipText(frame, plr)
  frame.text:SetText(Whisp:GetChatHistory(plr, Whisp.db.char.linesShown))
  frame.text:SetPoint("TOPLEFT",frame,"TOPLEFT",5,-5)
  -- Adjust height of the frame to the height of the text
  frame.text:SetWidth(frame:GetRight() - frame:GetLeft() - 10)
  frame:SetHeight(frame.text:GetHeight()+15)
end

-- Create
function Whisp:CreateEditBoxFrame()
  local frame = self:CreateMyFrame("EDITBOX")
  self:ResetEditBoxFrame(frame)
  frame:EnableMouse(true)
  -- Set draggable borders
  frame:SetResizable(true)
  frame.dragRight = self:CreateDrag(frame, "BOTTOMRIGHT", "BOTTOMRIGHT", "TOPRIGHT", "TOPRIGHT", "RIGHT")
  frame.dragLeft  = self:CreateDrag(frame, "BOTTOMLEFT", "BOTTOMLEFT", "TOPLEFT", "TOPLEFT", "LEFT")
  -- Set movement scripts
  frame:SetMovable(true)
  frame:SetScript("OnMouseDown", function() 
      if ((not Whisp.db.char.locked) and (arg1 == "LeftButton")) then
        this:StartMoving()
        this.isMoving = true
      elseif arg1=="RightButton" then  -- Show rightclick menu on the editbox tooltip
        if frame:GetLeft() < UIParent:GetRight()/2 then 
          Whisp:OpenMenu(Whisp.editBoxFrame.dragRight)
        else
          Whisp:OpenMenu(Whisp.editBoxFrame.dragLeft)
        end
      end
    end
  )
  frame:SetScript("OnMouseUp", function() 
      if ((not Whisp.db.char.locked)  and (arg1 == "LeftButton")) then
        this:StopMovingOrSizing()
        this.isMoving = false
      end
    end
  )
  frame:Hide()
  return frame
end

function Whisp:CreateTooltipFrame()
  local frame = self:CreateMyFrame("TOOLTIP")
  frame:Hide()
  return frame
end

function Whisp:CreateMyFrame(name)
  local frame = CreateFrame("Frame", "WHISP_"..name, UIParent)
  frame:ClearAllPoints()
  -- Size is required to be set or SetPoint() won't work later on
  frame:SetWidth(400)  
  frame:SetHeight(20)
  frame:SetMinResize(200,20)
  frame:SetFrameLevel(GameTooltip:GetFrameLevel() + 10)
  -- Set background and border
  frame:SetBackdrop({
    bgFile = "Interface\\ChatFrame\\ChatFrameBackground", tile = true, tileSize = 16,
    edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border", edgeSize = 16,
    insets = {left = 3, right = 3, top = 3, bottom = 3},
  })
  frame:SetFrameStrata("TOOLTIP")
  Whisp:UpdateBackgroundColor(frame)
  -- Setup textfields
  frame.text = frame:CreateFontString("WHISP_EDITBOXTEXT","OVERLAY","ChatFontNormal")
  frame.text:SetJustifyH("LEFT")
  frame.text:SetJustifyV("TOP")
  frame.text:SetTextColor(1.0,1.0,1.0,1.0)
  return frame
end


-- Setup draggable borders
function Whisp:CreateDrag(parent, anchor1, align1, anchor2, align2, direction)
  frame = CreateFrame("Frame", nil, parent)
  frame:Show()
  frame:SetFrameLevel(parent:GetFrameLevel() + 10)
  frame:SetWidth(16)
  frame:SetPoint(anchor1, parent, align1, 0, 0)
  frame:SetPoint(anchor2, parent, align2, 0, 0)
  frame:EnableMouse(true)
  frame:SetScript("OnMouseDown", function() 
      if ((not Whisp.db.char.locked) and (arg1 == "LeftButton")) then
        this:GetParent().isResizing = true
        this:GetParent():StartSizing(direction)
      end 
    end 
  )
  frame:SetScript("OnMouseUp", function() 
      if this:GetParent().isResizing == true then 
        this:GetParent():StopMovingOrSizing()
        this:GetParent().isResizing = false
        Whisp:ShowEditBoxFrame(Whisp.tellTarget) 
      end 
    end 
  )
  return frame
end

-- Various functions: Reset, background color and framelock
function Whisp:ResetEditBoxFrame(frame)
  if not frame then frame = Whisp.editBoxFrame end
  frame:ClearAllPoints()
  frame:SetWidth(ChatFrameEditBox:GetRight()-ChatFrameEditBox:GetLeft())
  if self.db.char.growUp then
    frame:SetPoint("BOTTOMLEFT",ChatFrameEditBox,"TOPLEFT")
  else
    frame:SetPoint("TOPLEFT",ChatFrameEditBox,"BOTTOMLEFT")
  end
end

function Whisp:UpdateBackgroundColor(frame)
  local c = Whisp.db.char.colorBackground
  frame:SetBackdropBorderColor(0.5, 0.5, 0.5, c[4])
  frame:SetBackdropColor(c[1], c[2], c[3], c[4])
end

function Whisp:ToggleFrameLock()
  if Whisp.db.char.locked then
    Whisp.editBoxFrame.dragLeft:Hide()
    Whisp.editBoxFrame.dragRight:Hide()
    Whisp.editBoxFrame:EnableMouse(false)
  else
    Whisp.editBoxFrame.dragLeft:Show()
    Whisp.editBoxFrame.dragRight:Show()
    Whisp.editBoxFrame:EnableMouse(true)
  end
end

-----------------------------------
-----------------------------------
-- FuBar tooltip functions

local tablet = AceLibrary("Tablet-2.0")
Whisp.tablet = tablet
function Whisp:OnTooltipUpdate()
    local cat = tablet:AddCategory('columns', 2)
    
    local contacts = {}
    for i,v in pairs(Whisp.chatHistory) do
      tinsert(contacts, {time = v.time[getn(v.time)], plr = i, inc = v.incoming[getn(v.incoming)]})
    end
    
    table.sort(contacts, function(a,b) return a.time>b.time end)
        
    for i=1, getn(contacts),1 do
      local inc = contacts[i].inc
      local plr = contacts[i].plr
      local time = contacts[i].time
      local color = inc and Whisp.db.profile.colorIncoming or Whisp.db.profile.colorOutgoing
      
      cat:AddLine(
          'text', time,
          'text2', plr,
          'text2R', color[1],
          'text2G', color[2],
          'text2B', color[3],
          'func', function(plr) 
            if IsControlKeyDown() then
              Whisp:ShowLogFrame(plr)
            else
              ChatFrameEditBox:Show()
              ChatFrameEditBox:SetText("/w "..plr.." ")
              ChatEdit_ParseText(ChatFrameEditBox, 0)
            end
          end,
          'arg1', plr,
          'onEnterFunc', function(plr) Whisp:ShowTooltipFrame(plr) end,
          'onEnterArg1', plr,
          'onLeaveFunc', function() Whisp.tooltipFrame:Hide() end,
          'onLeaveArg1', plr
      )    
    end
    
    tablet:SetHint(L["Click to reply\nControl-click to open log"])
end

function Whisp:OnTextUpdate()
  if self.lastSender and self.chatHistory[self.lastSender] then
    local v = self.chatHistory[self.lastSender]
    local inc = v.incoming[getn(v.incoming)]
    local color = inc and self.db.profile.colorIncoming or self.db.profile.colorOutgoing
    self:SetText(self:Colorise(self.lastSender, self:HexColor(color)))
  else
    self:SetText(self.lastSender and self.lastSender or "Whisp")  
  end
end

function Whisp:OnClick(button)
  if self.lastSender then
    if IsControlKeyDown() then
      Whisp:ShowLogFrame(self.lastSender)
    else
      ChatFrameEditBox:Show()
      ChatFrameEditBox:SetText("/w "..self.lastSender.." ")
      ChatEdit_ParseText(ChatFrameEditBox, 0)
    end
  end
end

-----------------------------------
-----------------------------------
-- Log window functions

function Whisp:ShowLogFrame(plr)
  if plr and Whisp.chatHistory[plr] then
    local msg = Whisp:GetChatHistory(plr, getn(Whisp.chatHistory[plr].time))
    if msg then
      Whisp_EditBox:SetText(msg)
      WhispUIParent:Show()
    end
  end
end

-----------------------------------
-----------------------------------
-- ChatHistory

-- Chat message event
function Whisp:UpdateChatHistory(msg, plr, snd, incoming)
  if not plr then return end
  plr = Whisp:ToName(plr)
  snd = Whisp:ToName(plr)
  Whisp.lastSender = plr
  -- Insert new message
  if not Whisp.chatHistory[plr] then
    Whisp.chatHistory[plr] = {sender = {}, message = {}, time = {}, incoming = {}, systime = {}}
  end
  tinsert(Whisp.chatHistory[plr].sender, snd)
  tinsert(Whisp.chatHistory[plr].message, msg)
  tinsert(Whisp.chatHistory[plr].time, date("%H:%M:%S"))
  tinsert(Whisp.chatHistory[plr].incoming, incoming)
  tinsert(Whisp.chatHistory[plr].systime, GetTime())
end

-- Returns the combined recent messages from this player
function Whisp:GetChatHistory(plr, max)
  local msg = ""
  if plr then
    plr = strupper(strsub(plr, 1,1))..strlower(strsub(plr,2))
    if Whisp.chatHistory[plr] then
      local n = getn(Whisp.chatHistory[plr].message)
      for i= n>max and n-max or 1, n, 1 do
        if i>1 and not (i==n-max) then 
          msg = msg.."\n"
        end
        local m = Whisp.chatHistory[plr].message[i]
        local s = Whisp.chatHistory[plr].sender[i]
        local t = Whisp.chatHistory[plr].time[i]
        local inc= Whisp.chatHistory[plr].incoming[i]
        local color = inc and Whisp.db.profile.colorIncoming or Whisp.db.profile.colorOutgoing
        m = Whisp:Colorise(m, Whisp:HexColor(color))
        s = Whisp:PlayerLink(s)
        msg = msg..t.." "..s..": "..m
      end
    end
  end
  if msg == "" then return nil end
  return msg
end

-- Garbage collection of message cache
function Whisp:CollectGarbage()
  -- Remove full entry of the player if the conversation has timed out
  local t = GetTime() - 60 * Whisp.db.profile.timeout
  for i=1, getn(Whisp.chatHistory), 1 do
    if Whisp.chatHistory[i].time[getn(Whisp.chatHistory[i].systime)] < t then
      tremove(Whisp.chatHistory,i)
    end
  end
  -- Remove individual entries of the player if they have timed out
  for i=1, getn(Whisp.chatHistory), 1 do
    for j=1, getn(Whisp.chatHistory[i].time) do
      if Whisp.chatHistory[i].systime[j] < t then
        tremove(Whisp.chatHistory.sender,j)
        tremove(Whisp.chatHistory.message,j)
        tremove(Whisp.chatHistory.time,j)
        tremove(Whisp.chatHistory.incoming,j)
        tremove(Whisp.chatHistory.systime,j)
      end
    end
  end
end

-----------------------------------
-----------------------------------
-- Utility and other functions

-- Output message to the chatframe
-- Level can be either:
-- 0: System message
-- 3: Debug
function Whisp:Msg(level, text)
  if not text then return end
  if level==0
  or level==3 and self.db.profile.notifyDebug then
    DEFAULT_CHAT_FRAME:AddMessage(text, 0.6, 0.6, 1)
  end
end

-- Create a link to a playername
function Whisp:PlayerLink(link)
  local type = "player:"..link
  return "|H"..type.."|h|cffffff00["..link.."]|r|h"
end

-- Returns the hexvalue of the rgb
function Whisp:HexColor(a1, a2, a3)
  local r, g, b = a1, a2, a3
  if type(a1)=="table" then
    r, g, b = a1[1], a1[2], a1[3]
  end
  return string.format("%02x%02x%02x", 255*r, 255*g, 255*b)
end

-- Colorise the text
function Whisp:Colorise(text, hexcolor)
  text = gsub(text, "(|c.-|H.-|r)", "|r%1|cff"..hexcolor)
  return "|cff"..hexcolor..text.."|r"
end

-- Format a name to capital-lower format
function Whisp:ToName(plr)
  if plr then plr = strupper(strsub(plr, 1,1))..strlower(strsub(plr,2)) end
  return plr
end

-- Functionality for versioning
function Whisp:CheckVersion()
  if not self.db.profile.version == self.currentBuild then
    self:Msg(0, format(L["Whisp v%s by Anea. Type |cffffffff/whisp|r or right-click the icon for options."], "|cffffffff"..self.db.profile.version.."|r"))
  end
  self.db.profile.version = self.currentBuild
end
